/*
 * This program is original work by San Mehat <nettwerk@valinux.com> and
 * Zac Sprackett <zac@sprackett.com>.
 * This program is distributed under the GNU Lesser General Public License
 * (LGPL).
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2.1 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *
 */

#include <stdio.h>
#include <sys/socket.h>
#include <signal.h>
#include <netinet/in.h>
#include <sys/un.h>
#include <stdlib.h>
#include <sys/time.h>
#include <netdb.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <vacmclient_api.h>
#include <config.h>

#ifdef OPENSSL
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>
#endif

#include <nexxus_connection.h>
static NEXXUS_CONNECTION	*nexxus_head = NULL;
static NEXXUS_CONNECTION	*nexxus_tail = NULL;

#ifdef OPENSSL
static SSL_CTX						*ctx = NULL;
static SSL_METHOD					*meth = NULL;
#endif

int	api_nexxus_suppress_error_messages = 0;

int			vacm_errno = 0;
#define API_FUNC_RETURN(x) { vacm_errno = x; return(x);}

/*
 * Internal function to quickly tell if a handle is valid or not
 */
static int handle_is_valid(void *handle)
{
	if (((NEXXUS_CONNECTION *) handle)->structure_key == 0xdeadbeef)
		return(1);
	return(0);
}

/*
 * Disconnect from the Nexxus pointed to by 'handle'
 */
int api_nexxus_disconnect(void *handle)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;

	if (!handle_is_valid(handle))
		{
		printf("api_nexxus_disconnect(): Bad nexxus handle\n");
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
		}

#ifdef OPENSSL
	SSL_shutdown(nexxus->ssl);
	SSL_free(nexxus->ssl);
#endif
	close(nexxus->I_fd);
  if (nexxus == nexxus_head)
    {
    if (nexxus_head->next)
      nexxus_head->next->prev = NULL;
    nexxus_head = nexxus->next;
    }
  else if (nexxus == nexxus_tail)
    {
    if (nexxus_tail->prev)
      nexxus_tail->prev->next = NULL;
    nexxus_tail = nexxus->prev;
    }
  else
    {
    nexxus->prev->next = nexxus->next;
    nexxus->next->prev = nexxus->prev;
    }
	memset(nexxus, 0, sizeof(NEXXUS_CONNECTION));
  free(nexxus);
	API_FUNC_RETURN(API_RETURN_OK)
}

/*
 * Lookup the address of 'addr', connect to Nexxus at that addr,
 * and put the handle in '*new_handle'
 */
int api_nexxus_connect(char *addr,
							 				 char *username,
											 char *password,
											 void **new_handle)
{
  struct hostent      *host;
  struct in_addr      inaddr;
  int                 I_tmpsock;
  int                 I_optval = 1;
  struct sockaddr_in  sockaddr;
  NEXXUS_CONNECTION   *nexxus_newcon;
	char								authmsg_out[255];
	char								*authmsg_in;
	__uint32_t					authlen;
	char								*bword;

#ifdef OPENSSL
	int									err;
	char								*str;

	if (!ctx)
		{
		SSLeay_add_ssl_algorithms();
		meth = SSLv23_client_method();
		SSL_load_error_strings();
		ctx = SSL_CTX_new(meth);
		}
#endif

	*new_handle = (void *) 0x00;
  if (!addr)
    API_FUNC_RETURN(API_RETURN_BAD_ADDRESS)
  host = gethostbyname(addr);
  if (!host)
    API_FUNC_RETURN(API_RETURN_LOOKUP_FAILED)
  memcpy(&inaddr, host->h_addr_list[0], sizeof(struct in_addr));

  nexxus_newcon = nexxus_head;
  while(nexxus_newcon)
    {
    if (nexxus_newcon->ip_addr.s_addr == inaddr.s_addr)
			API_FUNC_RETURN(API_RETURN_ALREADY_CONNECTED)
    nexxus_newcon = nexxus_newcon->next;
    }
  if ((I_tmpsock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("socket");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  if (setsockopt(I_tmpsock,
                 SOL_SOCKET,
                 SO_REUSEADDR,
                 &I_optval,
                 sizeof(I_optval)) <0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("setsockopt");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  memset(&sockaddr, 0, sizeof(sockaddr));
  sockaddr.sin_family = AF_INET;
  sockaddr.sin_addr.s_addr = inaddr.s_addr;
  sockaddr.sin_port = htons(6442);
  if (connect(I_tmpsock,
      (struct sockaddr *) &sockaddr,
      sizeof(struct sockaddr)) <0)
    {
    close(I_tmpsock);
	if (api_nexxus_suppress_error_messages == 0)
    perror("connect");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  if (!(nexxus_newcon = (NEXXUS_CONNECTION *) malloc(sizeof(NEXXUS_CONNECTION))))
		{
		printf("Unable to malloc nexxus connection structure (%m)\n");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}
  memset(nexxus_newcon, 0, sizeof(NEXXUS_CONNECTION));
  if (!nexxus_head)
    nexxus_head = nexxus_tail = nexxus_newcon;
  else
    {
    nexxus_tail->next = nexxus_newcon;
    nexxus_newcon->prev = nexxus_tail;
    nexxus_tail = nexxus_newcon;
    }
  nexxus_newcon->I_fd = I_tmpsock;
  memcpy(&nexxus_newcon->ip_addr, &inaddr, sizeof(inaddr));
	nexxus_newcon->structure_key = 0xdeadbeef;

#ifdef OPENSSL
	nexxus_newcon->ssl = SSL_new(ctx);
	SSL_set_fd(nexxus_newcon->ssl, nexxus_newcon->I_fd);
	err = SSL_connect(nexxus_newcon->ssl);
	if (!(nexxus_newcon->server_cert=SSL_get_peer_certificate(nexxus_newcon->ssl)))
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
	if ((str=X509_NAME_oneline(X509_get_subject_name(nexxus_newcon->server_cert),0,0)))
		free(str);
	if ((str=X509_NAME_oneline(X509_get_issuer_name(nexxus_newcon->server_cert),0,0)))
		free(str);
	X509_free(nexxus_newcon->server_cert);
#endif

	*new_handle = (void *) nexxus_newcon;

	/* Authenticate to nexxus */
	sprintf(authmsg_out,"NEXXUS:AUTH:%s:%s",username,password);
	if (api_nexxus_send_ipc(*new_handle,
													authmsg_out,
													(strlen(authmsg_out)+1))!=API_RETURN_OK)
		{
		goto misc_err;
		}
	memset(&authmsg_out[0], 0, sizeof(authmsg_out));

	/* Get the NEXXUS_READY message */
	if (api_nexxus_wait_for_data(*new_handle,
															 &authmsg_in,
															 &authlen,
															 10)!=API_RETURN_OK)
		{
		goto misc_err;
		}
	if (strcmp(authmsg_in,"NEXXUS_READY"))
		{
		free(authmsg_in);
		goto misc_err;
		}
	free(authmsg_in);
	/* Snarf the job started message */
	if (api_nexxus_wait_for_data(*new_handle,
															 &authmsg_in,
															 &authlen,
															 10)!=API_RETURN_OK)
		{
		goto misc_err;
		}
	free(authmsg_in);
	/* Snarf the authentication status */
	if (api_nexxus_wait_for_data(*new_handle,
															 &authmsg_in,
															 &authlen,
															 10)!=API_RETURN_OK)
		{
		goto misc_err;
		}
	if (!(bword = strtok(authmsg_in,":")))
		{
		free(authmsg_in);
		goto misc_err;
		}
	if (strcmp(bword,"NEXXUS"))
		{
		free(authmsg_in);
		goto misc_err;
		}
	if (!(bword = strtok(NULL,":")))
		{
		free(authmsg_in);
		goto misc_err;
		}
	if (!(bword = strtok(NULL,":")))
		{
		free(authmsg_in);
		goto misc_err;
		}
	if (strcmp(bword,"JOB_COMPLETED"))
		{
		free(authmsg_in);
		goto auth_err;
		}
	free(authmsg_in);

	/* Now fudge it so nexxus 'resends' us the NEXXUS_READY message */
	strcpy(authmsg_out,"NEXXUS:echo:NEXXUS_READY");
	if (api_nexxus_send_ipc(*new_handle,
													authmsg_out,
													(strlen(authmsg_out)+1))!=API_RETURN_OK)
		{
		goto misc_err;
		}
	/* Snarf the job started message */
	if (api_nexxus_wait_for_data(*new_handle,
															 &authmsg_in,
															 &authlen,
															 10)!=API_RETURN_OK)
		{
		goto misc_err;
		}
	free(authmsg_in);
	/* next data on pipe will be the 'NEXXUS_READY' message */

	API_FUNC_RETURN(API_RETURN_OK)
auth_err:
	api_nexxus_disconnect(*new_handle);
	*new_handle = 0;
	API_FUNC_RETURN(API_RETURN_AUTH_FAILED)
misc_err:
	api_nexxus_disconnect(*new_handle);
	*new_handle = 0;
	API_FUNC_RETURN(API_RETURN_MISC_ERROR)
}

/*
 * Lookup the socket fd of 'handle' and put it in '*fd'
 */
int api_nexxus_return_fd(void *handle, int *fd)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;

	if (!handle_is_valid(handle))
		{
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
		}
	*fd = nexxus->I_fd;
	API_FUNC_RETURN(API_RETURN_OK)
}

/*
 * Fills in 'ip_addr' based on 'handle'
 */
int api_nexxus_return_ip(void *handle, struct in_addr *ip_addr)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;

	if (!handle_is_valid(handle))
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
	memcpy(ip_addr, &nexxus->ip_addr, sizeof(nexxus->ip_addr));
	API_FUNC_RETURN(API_RETURN_OK)
}

/*
 * Fills in 'handle' based on 'addr'
 */
int api_nexxus_return_handle(void **handle, char *addr)
{
  struct hostent      *host;
  struct in_addr      inaddr;
  NEXXUS_CONNECTION   *nexxus;

  if (!addr)
    API_FUNC_RETURN(API_RETURN_BAD_ADDRESS)
  host = gethostbyname(addr);
  if (!host)
    API_FUNC_RETURN(API_RETURN_LOOKUP_FAILED)
  memcpy(&inaddr, host->h_addr_list[0], sizeof(struct in_addr));
	nexxus = nexxus_head;
	while(nexxus)
		{
    if (nexxus->ip_addr.s_addr == inaddr.s_addr)
			{
			*handle = (void *) nexxus;
			API_FUNC_RETURN(API_RETURN_OK)
		
			}
		nexxus = nexxus->next;
		}
	API_FUNC_RETURN(API_RETURN_LOOKUP_FAILED)
}

/*
 * Send a msg in 'pkt' of size 'len' bytes to Nexxus 'handle'
 */
int api_nexxus_send_ipc(void *handle, char *pkt, __uint32_t	len)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;
	__uint32_t				network_len;
#ifdef OPENSSL
	int								err;
#endif

	if (!handle_is_valid(handle))
		{
		printf("api_nexxus_send_ipc(): Bad nexxus handle\n");
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
		}

	network_len = htonl(len);
#ifdef OPENSSL
	err = SSL_write(nexxus->ssl, (char *) &network_len, sizeof(network_len));
	err = SSL_write(nexxus->ssl, &pkt[0], (int) len);
#else
  if (write(nexxus->I_fd, &network_len, sizeof(network_len))<0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("write length");
    API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  if (write(nexxus->I_fd, &pkt[0], len)<0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("write data");
    API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
#endif
  API_FUNC_RETURN(API_RETURN_OK)
}


/*
 * Read IPC message from Nexxus. Callers responsibility to free buffer
 */
int api_nexxus_recv_ipc(void *handle, char **buffer, __uint32_t	*len)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;
	__uint32_t				brtr;
	char							*offset;
	int								rc;

	if (!handle_is_valid(handle))
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
#ifdef OPENSSL
	if (!(rc = SSL_read(nexxus->ssl, (char *) len, sizeof(len))))
		API_FUNC_RETURN(API_RETURN_NEXXUS_DISCONNECT)
	*len = ntohl(*len);
	if (*len > 512)
		{
		printf("Packet corrupted or packet too long (len = %d (0x%x))\n",*len, *len);
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}
	*buffer = malloc(*len);
	if (*buffer == NULL) 
		{	
		printf("Unable to malloc %d bytes (%m)\n",*len);
		*len = htonl(*len);
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
		if (!(rc = SSL_read(nexxus->ssl, offset, (int) brtr)))
			API_FUNC_RETURN(API_RETURN_NEXXUS_DISCONNECT)
		else if (rc < 0)
			API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		brtr -= rc;
		offset += rc;
		}
#else
	if ((rc = read(nexxus->I_fd, len, sizeof(len)))<0)
		{
	if (api_nexxus_suppress_error_messages == 0)
    perror("read");
    API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}
	if (!rc)
		API_FUNC_RETURN(API_RETURN_NEXXUS_DISCONNECT)
	*len = ntohl(*len);
	if (*len > 512)
		{
		printf("Packet corrupted or packet too long (len = %d (0x%x))\n",*len,*len);
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}

	*buffer = malloc(*len);
	if (*buffer == NULL) {		/* bad malloc */
		printf("Unable to malloc %d bytes (%m)\n",*len);
		*len = htonl(*len);
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
	}
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr)
		{
		if ((rc = read(nexxus->I_fd, offset, (size_t) brtr))<0)
			{
			if (api_nexxus_suppress_error_messages == 0)
				perror("read");
			API_FUNC_RETURN(API_RETURN_MISC_ERROR)
			}
		if (!rc)
			API_FUNC_RETURN(API_RETURN_NEXXUS_DISCONNECT)
		brtr -= rc;
		offset += rc;
		}
#endif 
// 	printf("libvacm: read <%s>\n",*buffer);
	API_FUNC_RETURN(API_RETURN_OK)
}

int api_nexxus_wait_for_data(void				*handle,
														 char				**buffer,
														 __uint32_t	*len,
														 int				timeout)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;
  fd_set          	readfds;
  struct timeval  	to;
  int             	rc;

	if (!handle_is_valid(handle))
		{
		printf("api_nexxus_recv_ipc(): Bad nexxus handle\n");
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
		}
  to.tv_usec =0;
  to.tv_sec  =timeout;

  FD_ZERO(&readfds);
  FD_SET(nexxus->I_fd, &readfds);
  if ((rc =select((nexxus->I_fd+1), &readfds, NULL, NULL, &to)) < 0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("select");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  if (rc == 0)
		API_FUNC_RETURN(API_RETURN_TIMED_OUT)
	return(api_nexxus_recv_ipc(handle, buffer, len));
}

int api_nexxus_wait_short_for_data(void				*handle,
														 char				**buffer,
														 __uint32_t	*len,
														 int				timeout)
{
	NEXXUS_CONNECTION	*nexxus = (NEXXUS_CONNECTION *) handle;
  fd_set          	readfds;
  struct timeval  	to;
  int             	rc;

	if (!handle_is_valid(handle))
		{
		printf("api_nexxus_wait_short_for_data(): Bad nexxus handle\n");
		API_FUNC_RETURN(API_RETURN_BAD_HANDLE)
		}
  to.tv_usec =timeout;
  to.tv_sec  =0;

  FD_ZERO(&readfds);
  FD_SET(nexxus->I_fd, &readfds);
  if ((rc =select((nexxus->I_fd+1), &readfds, NULL, NULL, &to)) < 0)
    {
	if (api_nexxus_suppress_error_messages == 0)
    perror("select");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  if (rc == 0)
		API_FUNC_RETURN(API_RETURN_TIMED_OUT)
	return(api_nexxus_recv_ipc(handle, buffer, len));
}

int api_nexxus_ping(struct in_addr *ip_addr, struct timeval *to)
{
	struct sockaddr_in	addr_sender, addr_receiver;
	int									fd_sender, fd_receiver, len, fromlen, rc;
	fd_set							read_fds;
	char								msg[16];

  /* Setup the ping sender */
  if ((fd_sender = socket(AF_INET, SOCK_DGRAM, 0))<0)
    {
		if (!api_nexxus_suppress_error_messages)
    	perror("socket");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  memset(&addr_sender, 0, sizeof(addr_sender));
  addr_sender.sin_family = AF_INET;
  addr_sender.sin_port = htons(6445);
	memcpy(&addr_sender.sin_addr, ip_addr, sizeof(struct in_addr));
  
  /* Setup the pong receiver */
  if ((fd_receiver = socket(AF_INET, SOCK_DGRAM, 0))<0)
    {
		if (!api_nexxus_suppress_error_messages)
    	perror("socket");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  memset(&addr_receiver, 0, sizeof(addr_receiver));
  addr_receiver.sin_family = AF_INET;
  addr_receiver.sin_port = htons(6446);
  addr_receiver.sin_addr.s_addr = htonl(INADDR_ANY);
  if (bind(fd_receiver,
           (struct sockaddr *) &addr_receiver,
           sizeof(addr_receiver))<0)
    {      
		close(fd_receiver);
		close(fd_sender);
		if (!api_nexxus_suppress_error_messages)
    	perror("bind");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  /* Do the ping */
  strcpy(msg,"PING");
  len = sizeof(addr_sender);
  if (sendto(fd_sender,
             (void *) &msg[0],
             (strlen(msg)+1),
             0,
             (struct sockaddr *) &addr_sender,
             len)<0)
    {        
		close(fd_receiver);
		close(fd_sender);
		if (!api_nexxus_suppress_error_messages)
    	perror("sendto");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  FD_ZERO(&read_fds);
  FD_SET(fd_receiver, &read_fds);

  if ((rc = select((fd_receiver+1), &read_fds, NULL, NULL, to))<0)
    {
		close(fd_receiver);
		close(fd_sender);
		if (!api_nexxus_suppress_error_messages)
    	perror("select");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
    }
  else if (!rc)
		{
		close(fd_receiver);
		close(fd_sender);
		API_FUNC_RETURN(API_RETURN_TIMED_OUT)
		}
  else if (FD_ISSET(fd_receiver, &read_fds))
    {
    fromlen = sizeof(struct sockaddr_in);
    if (recvfrom(fd_receiver,
                 (void *) &msg[0],
                 sizeof(msg),
                 0,
                 (struct sockaddr *) &addr_receiver,
                 &fromlen) < 0)
      {          
			close(fd_receiver);
			close(fd_sender);
			if (!api_nexxus_suppress_error_messages)
    		perror("recvfrom");
			API_FUNC_RETURN(API_RETURN_MISC_ERROR)
      }
    if (strncmp(msg,"PONG",4))
			{
			close(fd_receiver);
			close(fd_sender);
			API_FUNC_RETURN(API_RETURN_TIMED_OUT)
			}
		close(fd_receiver);
		close(fd_sender);
		API_FUNC_RETURN(API_RETURN_OK)
    }
	else
		{
		close(fd_receiver);
		close(fd_sender);
		if (!api_nexxus_suppress_error_messages)
			printf("select fired on unknown entity\n");
		API_FUNC_RETURN(API_RETURN_MISC_ERROR)
		}
}

void api_nexxus_perror(char *msg)
{
	switch(vacm_errno)
		{
		case API_RETURN_OK:
			printf("%s: API_RETURN_OK\n",msg);
			break;
		case API_RETURN_BAD_HANDLE:
			printf("%s: API_RETURN_BAD_HANDLE\n",msg);
			break;
		case API_RETURN_NEXXUS_DISCONNECT:
			printf("%s: API_RETURN_NEXXUS_DISCONNECT\n",msg);
			break;
		case API_RETURN_LOOKUP_FAILED:
			printf("%s: API_RETURN_LOOKUP_FAILED\n",msg);
			break;
		case API_RETURN_BAD_ADDRESS:
			printf("%s: API_RETURN_BAD_ADDRESS\n",msg);
			break;
		case API_RETURN_MISC_ERROR:
			printf("%s: API_RETURN_MISC_ERROR\n",msg);
			break;
		case API_RETURN_ALREADY_CONNECTED:
			printf("%s: API_RETURN_ALREADY_CONNECTED\n",msg);
			break;
		case API_RETURN_TIMED_OUT:
			printf("%s: API_RETURN_TIMED_OUT\n",msg);
			break;
		case API_RETURN_AUTH_FAILED:
			printf("%s: API_RETURN_AUTH_FAILED\n",msg);
			break;
		default:
			printf("%s: UNKNOWN_ERROR_CODE\n",msg);
     }
}

char *
api_nexxus_get_error(void)
{
	switch(vacm_errno)
		{
    case API_RETURN_OK:
			return("Ok");
		case API_RETURN_BAD_HANDLE:
			return("Bad Handle");
		case API_RETURN_NEXXUS_DISCONNECT:
			return("Nexxus Disconnected");
		case API_RETURN_LOOKUP_FAILED:
			return("Return Lookup Failed");
		case API_RETURN_BAD_ADDRESS:
			return("Bad Address");
		case API_RETURN_MISC_ERROR:
			return("Miscellaneous Error");
		case API_RETURN_ALREADY_CONNECTED:
			return("Already Connected");
		case API_RETURN_TIMED_OUT:
			return("Return Timed Out");
		case API_RETURN_AUTH_FAILED:
			return("Authentication Failed");
		default:
			return("Unknown Error");
		}
	return("Eh?");
}
