/*
 * Copyright (C) 2000 Silicon Graphics, Inc.  All Rights Reserved.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of version 2 of the GNU General Public License as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it would be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  Further, any license provided herein,
 * whether implied or otherwise, is limited to this program in accordance
 * with the express provisions of the GNU General Public License.  Patent
 * licenses, if any, provided herein do not apply to combinations of this
 * program with other product or programs, or any other product whatsoever.
 * This program is distributed without any warranty that the program is
 * delivered free of the rightful claim of any third person by way of
 * infringement or the like.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write the Free Software Foundation, Inc.,
 * 59 Temple Place - Suite 330, Boston MA 02111-1307, USA.
 *
 */

/*
 * Author:      Dean Johnson (dtj@sgi.com)
 *
 * Purpose:     Handle just about everything having to do with the
 *		slaves.
 *
 * This Module is based heavily on the other loose modules by San Mehat
 * <nettwerk@valinux.com> and Zac Sprackett <zsprackett@valinux.com>.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include <sys/un.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <libloose.h>
//#include <prototypes.h>
#include <config.h>
#include <msc_slave_context.h>
#include <msc_module_harness.h>

// #define SLAVE_DEBUG

SLAVE_CONTEXT	*context_head = NULL;
SLAVE_CONTEXT	*context_tail = NULL;

static SLAVE	*slave_head = NULL;
static SLAVE	*slave_tail = NULL;

extern NODELIST	*nodelist_head;
extern int	num_nodes_per_slave;

extern void	node_destroy(NODELIST *node);

static int	read_message(int, char **, __uint32_t *);
static void	process_msg_from_master(int, int);
static void	process_msg_from_thread(SLAVE_CONTEXT *, int, int);
static int	send_message(int,char *);
static void	context_free(SLAVE_CONTEXT *context);
static SLAVE_CONTEXT	*context_alloc(char *node_id);
static void 	context_free_resources(SLAVE_CONTEXT *ctx);
static int 	thread_send(SLAVE_CONTEXT *ctx, char *msg);
static SLAVE	*slave_create(void);
static SLAVE	*slave_alloc(void);
static void	slave_free(SLAVE *slave);
static void	slave_msg_handler(int fd, void *data);
static void	free_parent_resources(SLAVE *slave);

extern void	*thread_start(void *);

int	slave_assign_and_notify(NODELIST *);
int	slave_remove_node(NODELIST *);
SLAVE	*locate_available_slave(void);
int	slave_notify_variable(NODELIST *, char *, char *);
int	slave_notify_rename_node(NODELIST *, char *);
int	slave_notify_del_node(NODELIST *);
int	slave_notify_new_node(NODELIST *);
void	slave_destroy(SLAVE *);
int	slave_send(SLAVE *, char *);
int	slave_deinit(void);
int	slave_notify_disconnection(int client_fd);
void	slave_machine_start(int, int);

int	context_variable_set(char *, char *, char *);
int	context_rename(char *, char *);
int	context_send_ipc(char *, char *);
int	context_destroy(char *);
SLAVE_CONTEXT *context_create(char *);;
SLAVE_CONTEXT *context_lookup(char *);
void	context_destroy_all(void);
void	context_init(void);
void	context_notify_disconnection_all(int fd);

/*
 * Assign a node to a slave and notify the slave
 */

int 
slave_assign_and_notify(NODELIST *node)
{
	if (!(node->slave = locate_available_slave())) {
		if (!(node->slave = slave_create()))
			return(-ENOMEM);
	}
	node->slave->nodecount++;
	return(slave_notify_new_node(node));
}

/*
 * Notify a slave that a node is disapearing
 * Kill the slave if it has no nodes
 */

int 
slave_remove_node(NODELIST *node)
{
	int	error;

	node->slave->nodecount--;
	error = slave_notify_del_node(node);
	if (node->slave->nodecount == 0)
		slave_destroy(node->slave);
	node->slave = NULL;
	return(error);
}

/*
 * Find a slave with some free slots in it
 */

SLAVE *
locate_available_slave()
{
	SLAVE	*slave_ptr;

	for (slave_ptr = slave_head; slave_ptr; slave_ptr = slave_ptr->next) {
		if (slave_ptr->nodecount < num_nodes_per_slave)
			return(slave_ptr);
	}
	return(NULL);
}

/*
 * Notify a slave of a variable change 
 */

int 
slave_notify_variable(NODELIST *node, char *variable, char *value)
{
	char	msg[255];
	int		error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_VAR:%s:%s:%s", node->node_id, variable, value);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave that a client has disconnected
 */

int 
slave_notify_disconnection(int client_fd)
{
	char	msg[255];
	SLAVE	*slv;
	int	error;

	sprintf(msg,"CLIENT_DIS:NONE:%d",client_fd);
	for (slv = slave_head; slv; slv = slv->next) {
		if ((error = slave_send(slv, msg)) < 0)
			return(error);
	}
	return(0);
}

/*
 * Notify a slave that a node name has changed
 */

int 
slave_notify_rename_node(NODELIST *node, char *new_name)
{
	char	msg[255];
	int	error;

	sprintf(msg, "NODE_RENAME:%s:%s", node->node_id, new_name);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave that a node is gone (deleted)
 */

int 
slave_notify_del_node(NODELIST *node)
{
	char	msg[255];
	int	error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_DEL:%s", node->node_id);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify a slave of its new burden (node)
 * Returns 0 on success, negative errno on failure
 */

int 
slave_notify_new_node(NODELIST *node)
{
	char	msg[255];
	int	error;

	if (!node->slave)
		return(-EBADSLT);
	sprintf(msg, "NODE_NEW:%s", node->node_id);
	if ((error = slave_send(node->slave, msg)) < 0)
		return(error);
	return(0);
}

/*
 * Notify all slaves of shutdown
 */

int 
slave_deinit()
{
	SLAVE	*slv;
	char	slave_msg[255];

	strcpy(slave_msg,"SLAVE_SHUTDOWN:NONE:NONE");
	for (slv = slave_head; slv; slv = slv->next) {
		slave_send(slv, slave_msg);
	}
	return(0);
}

/*
 * Allocate a slave structure and fork off a slave
 */

static SLAVE *
slave_create()
{
	SLAVE	*new_slave;
	int	pipefd1[2];
	int	pipefd2[2];

	if ((pipe(pipefd1) < 0 ) || (pipe(pipefd2) < 0)) {
		lm_log("Unable to create thread pipes");
		return(NULL);
	}
	if (!(new_slave = slave_alloc()))
		return(NULL);
	new_slave->read_sock = pipefd1[0];
	new_slave->write_sock = pipefd2[1];
	if (!(new_slave->pid = fork())) {
		/* New slave */
		free_parent_resources(new_slave);
		slave_machine_start(pipefd2[0], pipefd1[1]);
		_exit(0);
	}
	new_slave->nodecount = 0;
	lm_watch_fd(new_slave->read_sock, slave_msg_handler, (void *) new_slave);
	return(new_slave);
}

/*
 * Called from within the child context
 * Frees up memory and closes file descriptors
 * that the child wont use (that were
 * inherited from the parent
 */

static void 
free_parent_resources(SLAVE *slave)
{
	SLAVE		*slave_scan;
	NODELIST	*node_scan;

	slave_scan = slave_head;
	while (slave_scan) {
		if (slave_scan != slave) {
			/* We don't want to close our own read/write fd's */
			close(slave_scan->read_sock);
			close(slave_scan->write_sock);
		}
		slave_free(slave_scan);
		slave_scan = slave_head;
	}

	node_scan = nodelist_head;
	while (node_scan) {
		node_destroy(node_scan);
		node_scan = nodelist_head;
	}
}

/*
 * Shut down a slave and free up it's structure
 */

void 
slave_destroy(SLAVE *slave)
{
	NODELIST	*nodes;
	char		msg[255];

	/* Ask the slave to shutdown cleanly */
	strcpy(msg,"SLAVE_SHUTDOWN:NONE:NONE");
	slave_send(slave, msg);
	sleep(1);

	/* Send the slave a signal */
	kill(slave->pid, SIGKILL);

	/* Clean up any nodes that were using this slave */
	nodes = nodelist_head;
	while(nodes) {
		if (nodes->slave == slave)
			nodes->slave = NULL;
		nodes = nodes->next;
	}

	/* Now free the resources */
	lm_unwatch_fd(slave->read_sock);
	close(slave->read_sock);
	close(slave->write_sock);
	slave_free(slave);
}

/*
 * Allocate a slave structure
 */
static SLAVE *
slave_alloc()
{
	SLAVE	*new;

	if (!(new = (SLAVE *) malloc(sizeof(SLAVE)))) {
		lm_log("Unable to malloc new slave structure");
		return(NULL);
	}
	memset(new, 0, sizeof(SLAVE));
	if (!slave_head) {
    		slave_head = new;
    	}
  	else	{
		slave_tail->next = new;
		new->prev = slave_tail;
    	}
	slave_tail = new;
	return(new);
}

/*
 * Free a slave structure
 */

static void 
slave_free(SLAVE *slave)
{
	if (slave == slave_head) {
		slave_head = slave_head->next;
		if (slave_head)
			slave_head->prev = NULL;
		else
			slave_tail = NULL;
	}
	else if (slave == slave_tail) {
		slave_tail = slave_tail->prev;
		if (slave_tail)
			slave_tail->next = NULL;
		else
			slave_head = NULL;
	}
	else {
		slave->next->prev = slave->prev;
		slave->prev->next = slave->next;
	}
	free(slave);
}

int 
slave_send(SLAVE	*slave, char *msg)
{
	__uint32_t	len;
	__uint32_t	net_len;
	int		rc;

#if 0
	printf("Sending msg [%s] to slave pid %d (fd %d)\n", msg, slave->pid,
		slave->write_sock);
#endif

	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write (slave->write_sock, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	if ((rc = write (slave->write_sock, (void *) msg, len)) < 0)
		return(rc);
	return(0);
}

static int 
slave_recv(int fd, char **buffer, __uint32_t *len)
{
	__uint32_t	net_len;
	__uint32_t	brtr;
	int		rc;
	char		*offset;

	if ((rc = read(fd, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	*len = ntohl(net_len);

	if (!(*buffer = malloc(*len)))
    		return(-ENOMEM);
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while (brtr) {
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		brtr -= rc;
		offset += rc;
	}
	return(0);
}

/*
 * Handler for messages comming from a slave to the master 
 */

static void 
slave_msg_handler(int fd, void *data)
{
	int		rc, len;
	char	*msg_buffer;

	KNOWN_UNUSED_PARAM(data);

	if ((rc = slave_recv(fd, &msg_buffer, &len)) < 0) {
		printf("Master unable to read msg from slave (%s)\n",
			strerror(-rc));
		return;
	}
	// printf("Master got msg (%s) from slave\n",msg_buffer);
	lm_send_to_nexxus(msg_buffer);
	free(msg_buffer);
}

/*----------------------------------------------------------------------*
 *				MSC SLAVE MACHINE			*
 *----------------------------------------------------------------------*/

void
slave_machine_start(int read_fd, int write_fd)
{
	struct timeval	to;
	pid_t		pid;
	fd_set		read_fds;
	int		max, rc;
	SLAVE_CONTEXT	*ctx;
#ifdef	SLAVE_DEBUG
	char		log_msg[255];
#endif	/*SLAVE_DEBUG*/

	pid = getpid();
#ifdef	SLAVE_DEBUG
	sprintf(log_msg,"0:0:LOG:Slave pid %d starting",pid);
	send_message(write_fd, log_msg);
#endif	/*SLAVE_DEBUG*/

	context_init();

	while(1) {
		FD_ZERO(&read_fds);
		to.tv_sec = 1;
		to.tv_usec = 0;

		FD_SET(read_fd, &read_fds);
		max = read_fd;

		for (ctx = context_head; ctx; ctx = ctx->next) {
			FD_SET(ctx->slave_ipc_read, &read_fds);
			if (max < ctx->slave_ipc_read)
				max = ctx->slave_ipc_read;
		}
		if ((rc = select((max+1), &read_fds, NULL, NULL, &to)) < 0) {
			printf("Slave %d select failed\n",pid);
			continue;
		}
		else if (!rc) {
			continue;
		}
		if (FD_ISSET(read_fd, &read_fds)) {
			process_msg_from_master(read_fd, write_fd);
			continue;
		}
		for (ctx = context_head; ctx; ctx = ctx->next) {
			if (FD_ISSET(ctx->slave_ipc_read, &read_fds))
				process_msg_from_thread(ctx, read_fd, write_fd);
		}
	}
}

static void
process_msg_from_master(int r_fd, int w_fd)
{
	char	*msg_buffer, *cmd, *node_id, *variable, *value, *new_name;
	int	rc;
	char	*saved_buffer;
	__uint32_t	len;
	KNOWN_UNUSED_PARAM(cmd);
	KNOWN_UNUSED_PARAM(w_fd);

	if ((rc = read_message(r_fd, &msg_buffer, &len)) < 0) {
		printf("Slave unable to read msg from thread (%s)\n",strerror((-rc)));
		return;
	}
	saved_buffer = malloc(len);
	strcpy(saved_buffer, msg_buffer);

//	printf("Slave got msg (%s) from master\n",msg_buffer);
	if (!(cmd = strtok(msg_buffer, ":")))
		goto out;
	if (!(node_id = strtok(NULL, ":")))
		goto out;
	if (!strcasecmp(cmd, "NODE_NEW")) {
		if (!context_create(node_id)) {
			printf("MSC Slave unable to create node context\n");
			goto out;
		}
	}
	else if (!strcasecmp(cmd, "NODE_VAR")) {
		if (!(variable = strtok(NULL,":")))
			goto out;
		if (!(value = strtok(NULL,":")))
			goto out;
		if ((rc = context_variable_set(node_id, variable, value)) < 0) {
			printf("MSC Slave unable to set context variable %s (%s)\n",
						 variable,
						 strerror(-rc));
			goto out;
		}
	}
	else if (!strcasecmp(cmd, "CLIENT_DIS")) {
		char	*client_fd;
		int	fd;

		if (!(client_fd = strtok(NULL,":")))
			goto out;
		fd = atoi(client_fd);
		context_notify_disconnection_all(fd);
		goto out;
	}
	else if (!strcasecmp(cmd, "NODE_RENAME")) {
		if (!(new_name = strtok(NULL,":")))
			goto out;
		if ((rc = context_rename(node_id, new_name)) < 0) {
			printf("MSC Slave unable to rename context %s (%s)\n",
			       node_id,strerror(-rc));
			goto out;
		}
	}
	else if (!strcasecmp(cmd, "NODE_DEL")) {
		if ((rc = context_destroy(node_id)) < 0) {
			printf("MSC Slave unable to destroy node context (%s)\n",strerror(-rc));
			goto out;
		}
	}
	else if (!strcasecmp(cmd, "NODE_IPC")) {
		if ((rc = context_send_ipc(node_id, saved_buffer)) < 0) {
			printf("MSC Slave unable to send IPC (%s)\n",strerror(-rc));
			goto out;
		}
	}
	else if (!strcasecmp(cmd,"SLAVE_SHUTDOWN")) {
		context_destroy_all();
		exit(0);
	}
	else
		printf("MSC Slave received unknown msg from master (%s)\n",cmd);
out:
	free(msg_buffer);
	free(saved_buffer);
}

static void
process_msg_from_thread(SLAVE_CONTEXT *ctx, int read_fd, int write_fd)
{
	char	*msg_buffer;
	int	rc;
	__uint32_t	len;
	KNOWN_UNUSED_PARAM(read_fd);

	if ((rc = read_message(ctx->slave_ipc_read, &msg_buffer, &len)) < 0) {
		printf("Slave unable to read msg from thread (%s)\n",strerror((-rc)));
		return;
	}
	// printf("Slave got msg (%s) from thread\n",msg_buffer);
	/* Forward message to master */
	send_message(write_fd, msg_buffer);
	free(msg_buffer);
}

static int
read_message(int fd, char **buffer, __uint32_t *len)
{
	__uint32_t	net_len;
	int	rc;
	__uint32_t	brtr;
	char	*offset;

	brtr = sizeof(net_len);
	offset = (char *) &net_len;
	while (brtr) {
//		printf("IPC length hunk read basep=%p, offset=%p, brtr=%d\n",
//			&net_len,offset,brtr);
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset += rc;
	}
	*len = ntohl(net_len);

//	printf("IPC length size =%d\n",*len);
	if (!(*buffer = malloc(*len))) {
		printf("Unable to allocate memory of size %d (0x%x)\n",*len,*len);
		return(-ENOMEM);
	}
	memset(*buffer, 0, *len);
	brtr = *len;
	offset = *buffer;
	while(brtr) {
//		printf("IPC data hunk read basep=%p, offset=%p, brtr=%d\n",
//			*buffer,offset,brtr);
		if ((rc = read(fd, (void *) offset, brtr)) < 0)
			return(rc);
		else if (!rc)
			return(-1);
		brtr -= rc;
		offset +=rc;
	}
//	printf("IPC read complete (msg %s)\n",*buffer);
	return(0);
}

static int
send_message(int fd, char *msg)
{
	__uint32_t  len;
	__uint32_t  net_len;
	int         rc;

	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write (fd, (void *) &net_len, sizeof(net_len))) < 0)
		return(rc);
	if ((rc = write (fd, (void *) msg, len)) < 0)
		return(rc);
	return(0);
}

/*----------------------------------------------------------------------*
 *			     SLAVE_CONTEXT STUFF			*
 *----------------------------------------------------------------------*/

/*
 * Now and then a thread may want to peek at another threads
 * context structure. This lock is in place to keep bad things
 * from happening
 */

pthread_mutex_t	context_mutex;

void 
context_init()
{
	pthread_mutex_t tmp_mutex = PTHREAD_MUTEX_INITIALIZER;

	memcpy(&context_mutex, &tmp_mutex, sizeof(pthread_mutex_t));
	context_head = NULL;
	context_tail = NULL;
	pthread_mutex_init(&context_mutex, NULL);
}

/*
 * Set a context variable and notify the thread
 */

int 
context_variable_set(char *node_id, char *variable, char *value)
{
	SLAVE_CONTEXT *ctx;
	char		thread_msg[255];

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	sprintf(thread_msg,"NODE_VAR:%s:%s:%s",node_id,variable,value);
	return(thread_send(ctx, thread_msg));
}

/*
 * Rename a node and notify the thread
 */

int 
context_rename(char *node_id, char *new_id)
{
	SLAVE_CONTEXT *ctx;
	char		thread_msg[255];

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	sprintf(thread_msg,"NODE_RENAME:%s:%s",node_id,new_id);
	return(thread_send(ctx, thread_msg));
}

/*
 * Send IPC to a node thread
 */

int 
context_send_ipc(char *node_id, char *ipc_msg)
{
	SLAVE_CONTEXT *ctx;

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	return(thread_send(ctx, ipc_msg));
}

/*
 * Cleanly destroy a node context
 */

int 
context_destroy(char *node_id)
{
	struct timeval	now;
	__uint32_t	expire;
	int		rc;
	SLAVE_CONTEXT	*ctx;

	if (!(ctx = context_lookup(node_id)))
		return(-ESRCH);
	if (pthread_cancel(ctx->thread) < 0)
		return(-EBUSY);

	gettimeofday(&now, NULL);
	expire = now.tv_sec + 5;

	rc = 0;
	while (ctx->refcount && rc != ETIMEDOUT) {
		usleep(500);
		gettimeofday(&now, NULL);
		if (now.tv_sec > (int)expire)
			rc = ETIMEDOUT;
	}
	if (rc == ETIMEDOUT) {
		printf("MSC Slave thread for node %s not responding to cancel request\n",
			ctx->node_id);
		return(-ETIMEDOUT);
	}
	context_free(ctx);
	return(0);
}

/*
 * Notify all threads about a clients disconnection
 */

void 
context_notify_disconnection_all(int fd)
{
	SLAVE_CONTEXT	*ctx;
	char	msg[255];

	sprintf(msg,"CLIENT_DIS:NONE:%d",fd);
	for (ctx = context_head; ctx; ctx = ctx->next) {
		thread_send(ctx,msg);
	}
}

/*
 * Create a context for a node and start it's thread
 */

SLAVE_CONTEXT *
context_create(char *node_id)
{
	SLAVE_CONTEXT	*ctx;
	int	pipefd1[2], pipefd2[2];

	if (!(ctx = context_alloc(node_id)))
		return(NULL);
	if ((pipe(pipefd1) < 0) || (pipe(pipefd2) < 0)) {
		context_free(ctx);
		return(NULL);
	}
	ctx->thread_ipc_read  = pipefd2[0];
	ctx->thread_ipc_write = pipefd1[1];
	ctx->slave_ipc_read   = pipefd1[0];
	ctx->slave_ipc_write  = pipefd2[1];
	ctx->msc_fd 	      = -1;

	if (pthread_create(&ctx->thread, NULL, thread_start, (void *) ctx) < 0) {
		printf("MSC Slave pthread_create() call failed for node %s\n",
		       node_id);
		context_free(ctx);
		return(NULL);
	}
	return(ctx);
}

void 
context_destroy_all()
{
	SLAVE_CONTEXT	*ctx;
	int	rc;

	for (ctx = context_head; ctx; ctx = context_head) {
		if ((rc = context_destroy(ctx->node_id)) < 0) {
			printf("Slave unable to destroy thread for node %s (%s)\n",
						 ctx->node_id,
						 strerror(-rc));
		}
		// ctx = ctx -> next;
	}
}

SLAVE_CONTEXT *
context_lookup(char *node_id)
{
	SLAVE_CONTEXT	*ctx;

	ctx = context_head;
	for (ctx = context_head; ctx; ctx = ctx->next) {
		if (!strcasecmp(node_id, ctx->node_id))
			return(ctx);
	}
	return(NULL);
}

/*
 * Free any remaining resources for a context
 * This is called *after* the thread is cancelled
 */

static void
context_free_resources(SLAVE_CONTEXT *ctx)
{
	close(ctx->thread_ipc_read);
	close(ctx->thread_ipc_write);
	close(ctx->slave_ipc_read);
	close(ctx->slave_ipc_write);
	ctx->thread_ipc_read  = -1;
	ctx->thread_ipc_write = -1;
	ctx->slave_ipc_read   = -1;
	ctx->slave_ipc_write  = -1;
	if (ctx->node_id) {
		free(ctx->node_id);
		ctx->node_id = NULL;
	}
	if (ctx->addr) {
		free(ctx->addr);
		ctx->addr = NULL;
	}
}

static SLAVE_CONTEXT *
context_alloc(char *node_id)
{
	SLAVE_CONTEXT	*new;

	if (!node_id) {
		lm_log("NULL passed in as node_id for new context");
		return(NULL);
	}
	if (!(new = (SLAVE_CONTEXT *) malloc(sizeof(SLAVE_CONTEXT)))) {
		lm_log("Unable to malloc new context structure");
		return(NULL);
	}
	memset(new, 0, sizeof(SLAVE_CONTEXT));
	
	pthread_mutex_lock(&context_mutex);

	new->thread_ipc_read  = -1;
	new->thread_ipc_write = -1;
	new->slave_ipc_read   = -1;
	new->slave_ipc_write  = -1;
	new->con_sock = -1;
	if (!context_head) {
		context_head = new;
		context_tail = new;
    	}
  	else {
		context_tail->next = new;
		new->prev = context_tail;
		context_tail = new;
    	}
	new->node_id = malloc((strlen(node_id)+1));
	strcpy(new->node_id, node_id);

	pthread_mutex_unlock(&context_mutex);

	return(new);
}

static void
context_free(SLAVE_CONTEXT *context)
{
	pthread_mutex_lock(&context_mutex);

	if (context == context_head) {
    		context_head = context_head->next;
    		if (context_head)
      			context_head->prev = NULL;
    	}
  	else if (context == context_tail) {
		context_tail = context_tail->prev;
		if (context_tail)
			context_tail->next = NULL;
    	}
	else {
		context->next->prev = context->prev;
		context->prev->next = context->next;
    	}
	context_free_resources(context);

	pthread_mutex_unlock(&context_mutex);

	free(context);
}

static int
thread_send(SLAVE_CONTEXT *ctx, char *msg)
{
	__uint32_t  len;
	__uint32_t  net_len;
	int         rc;

	len = strlen(msg)+1;
	net_len = htonl(len);
	if ((rc = write(ctx->slave_ipc_write,(void *) &net_len,sizeof(net_len))) < 0)
		return(rc);
	if ((rc = write (ctx->slave_ipc_write, (void *) msg, len)) < 0)
		return(rc);
	return(0);
}
